package com.vcarecity.publish.elastic.service.impl;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.vcarecity.elastic.util.AgencyPathUtil;
import com.vcarecity.elastic.util.SnowFlake;
import com.vcarecity.publish.constants.ElasticConstant;
import com.vcarecity.publish.elastic.entity.UnitAgencyDetailEntity;
import com.vcarecity.publish.elastic.service.ElasticUpdateService;
import com.vcarecity.publish.entity.UnitAgencyMergeEntity;
import com.vcarecity.publish.sql.service.AgencyService;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.action.bulk.BulkRequest;
import org.elasticsearch.action.bulk.BulkResponse;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.common.xcontent.XContentType;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.reindex.BulkByScrollResponse;
import org.elasticsearch.index.reindex.DeleteByQueryRequest;
import org.elasticsearch.index.reindex.UpdateByQueryRequest;
import org.elasticsearch.script.Script;
import org.elasticsearch.script.ScriptType;
import org.springframework.cglib.beans.BeanCopier;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Kerry on 12/17/19
 */

@Slf4j
@Service
public class ElasticUpdateServiceImpl implements ElasticUpdateService {

    public final String index;

    private final RestHighLevelClient restHighLevelClient;
    private final ObjectMapper objectMapper;
    private final AgencyService agencyService;

    public ElasticUpdateServiceImpl(RestHighLevelClient restHighLevelClient,
                                    ObjectMapper objectMapper,
                                    AgencyService agencyService) {
        this.restHighLevelClient = restHighLevelClient;
        this.objectMapper = objectMapper;
        this.agencyService = agencyService;

        this.index = ElasticConstant.UNIT_AGENCY_INDEX;
    }


    @Override
    public void updateTest() throws IOException {

        Map<String, Object> params = new HashMap<>();
        params.put("dtuNo", "000000000001");
        params.put("puserCode", "/0/000000000001/1/12345678/");

        UpdateByQueryRequest request = new UpdateByQueryRequest("unit-agency");
        request.setQuery(QueryBuilders.termQuery("unitId", 12345678));
        request.setMaxDocs(1);
        request.setBatchSize(1);
        request.setScript(new Script(ScriptType.INLINE, "painless", getPainlessCode(params), params));
        request.setRefresh(true);


        BulkByScrollResponse response = restHighLevelClient.updateByQuery(request, RequestOptions.DEFAULT);

        System.out.println(response.getStatus());

    }

    @Override
    public void handlerDataChange(String key, List<UnitAgencyMergeEntity> data) throws IOException {
        List<UnitAgencyMergeEntity> updateList = new ArrayList<>();
        List<Long> deleteId = new ArrayList<>();
        for (UnitAgencyMergeEntity uae : data) {
            if (uae.getIsDeleted() == 1) {
                deleteId.add(uae.getUnitId());
            } else if (uae.getIsDeleted() == 0) {
                updateList.add(uae);
            }
        }

        logger.info("key = {}, update = {}, delete = {}", key, updateList.size(), deleteId.size());
        logger.debug("delete; {} : {}", key, deleteId);

        // 先删除
        if (!deleteId.isEmpty()) {
            this.deleteByQuery(key, deleteId);
        }

        if (updateList.isEmpty()) {
            return;
        }
        // 更新数据
        List<UnitAgencyMergeEntity> addData = updateDataByQuery(key, updateList);

        logger.info("add data size = {}", addData.size());

        // 添加数据
        this.addElasticData(addData);
    }


    @Override
    public void deleteByQuery(String key, List<Long> ids) throws IOException {
        if (ids == null || ids.isEmpty()) {
            return;
        }
        for (Long id : ids) {
            DeleteByQueryRequest request = new DeleteByQueryRequest(index);
            request.setQuery(QueryBuilders.termQuery(key, id));
            restHighLevelClient.deleteByQuery(request, RequestOptions.DEFAULT);
        }
    }

    @Override
    public List<UnitAgencyMergeEntity> updateDataByQuery(String key, List<UnitAgencyMergeEntity> data) throws IOException {
        List<UnitAgencyMergeEntity> addList = new ArrayList<>();
        final BeanCopier beanCopier = BeanCopier.create(UnitAgencyMergeEntity.class, UnitAgencyDetailEntity.class, false);
        for (UnitAgencyMergeEntity item : data) {

            Map<String, Object> params = createElasticDataMap(beanCopier, item);
            Object value = params.get(key);
            params.remove(key);

            UpdateByQueryRequest request = new UpdateByQueryRequest(index);
            request.setQuery(QueryBuilders.termQuery(key, value));
            request.setScript(new Script(ScriptType.INLINE, "painless", getPainlessCode(params), params));

            BulkByScrollResponse response = restHighLevelClient.updateByQuery(request, RequestOptions.DEFAULT);

            logger.info("response:{}", response);

            if (response.getUpdated() == 0L) {
                addList.add(item);
            }
        }

        return addList;
    }

    @Override
    public void addElasticData(List<UnitAgencyMergeEntity> data) throws IOException {
        if (data.isEmpty()) {
            return;
        }
        final BeanCopier beanCopier = BeanCopier.create(UnitAgencyMergeEntity.class, UnitAgencyDetailEntity.class, false);

        List<Map<String, Object>> saveResult = new ArrayList<>(data.size());

        for (UnitAgencyMergeEntity item : data) {
            Map<String, Object> map = createElasticDataMap(beanCopier, item);
            saveResult.add(map);
        }
        if (saveResult.isEmpty()) {
            return;
        }
        try {
            BulkRequest bulkRequest = new BulkRequest();
            for (Map<String, Object> val : saveResult) {
                IndexRequest indexRequest = createIndexRequest(index, val);
                bulkRequest.add(indexRequest);
            }
            BulkResponse bulkResponse = restHighLevelClient.bulk(bulkRequest, RequestOptions.DEFAULT);

            logger.debug("bulk addElasticData status {}, size = {}", bulkResponse.status(), saveResult.size());

        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private Map<String, Object> createElasticDataMap(BeanCopier beanCopier, UnitAgencyMergeEntity item) {
        List<Integer> agencyIds = AgencyPathUtil.splitAgencyId(item.getAgencyPath());
        List<String> agencyPathDetail = agencyService.getAgencyPathDetail(agencyIds);
        UnitAgencyDetailEntity target = new UnitAgencyDetailEntity();
        beanCopier.copy(item, target, null);
        target.setAgencyDetail(String.join("", agencyPathDetail));
        Map<String, Object> map = objectMapper.convertValue(target, new TypeReference<Map<String, Object>>() {
            @Override
            public Type getType() {
                return super.getType();
            }
        });
        final Map<String, Integer> queryMap = AgencyPathUtil.agencyPathQueryMap(agencyIds);
        map.putAll(queryMap);
        return map;
    }


    private IndexRequest createIndexRequest(String index, Map<String, Object> value) {
        IndexRequest indexRequest = new IndexRequest(index);
        indexRequest.id(SnowFlake.nextId() + "");
        indexRequest.source(value, XContentType.JSON);
        return indexRequest;
    }


    private String getPainlessCode(Map<String, Object> params) {
        String[] items = new String[params.size()];
        int index = 0;
        for (String key : params.keySet()) {
            items[index++] = "ctx._source." + key + "=params." + key;
        }
        return String.join(";", items);
    }
}
